/*
 * msf47.c --  the MSF module for WVLAN47, the WaveLAN/IEEE driver for Linux
 *
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define __NO_VERSION__ /* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>

char kernel_version [] = UTS_RELEASE;

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */
#include <linux/ioport.h>	/* check_region, request_region */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/tcp.h>         /* struct tcphdr */
#include <linux/skbuff.h>

#include <asm/checksum.h>

#include "snull.h"
#include "hcf.h"		//implies #include "../hcf/mdd.h"
#include "msf47.h"

#define PCIC_PORT	0x03E0
#define PCIC_RANGE	0x0002

#if !defined  HCF_PORT_0
#define HCF_PORT_0	0x0000
#endif
#if !defined  HCF_ACT_TICK_ON
#define HCF_ACT_TICK_ON	0xFFFF
#define HCF_ACT_TICK_OFF	0xFFFF
#endif

int	wvlan_change_mtu(struct device *dev, int new_mtu);
int	wvlan_config(struct device *dev, struct ifmap *map);
int	wvlan_init(struct device *dev);
int	wvlan_ioctl(struct device *dev, struct ifreq *rq, int cmd);
void	wvlan_isr(int irq, void *dev_id, struct pt_regs *regs);
int	wvlan_open(struct device *dev);
//int	wvlan_rebuild_header(void *buff, struct device *dev, unsigned long dst,
//		    struct sk_buff *skb);
int	wvlan_release(struct device *dev);
void	wvlan_rx(struct device *dev, int len);
struct enet_statistics *wvlan_stats(struct device *dev);
int	wvlan_tx(struct sk_buff *skb, struct device *dev);

/*;? do it hard coded till we figure out how to do it with the PCMCIA deamon */
void pcic_pair_out( hcf_16 reg, hcf_8 val );
void pcic_setup( void );

/* This is a load-time options */
static int eth = 0;		/* Call yourself "eth#". Default is "wvlan47#"  	*/
static int port = 0x1200;	/* default to io_base of 1200, cmd line override	*/
static int irq = 0x0a;		/* default to irq of 10, cmd line override		*/
static int port_type = 0x01;	/* default to ESS					*/	
static int verb = 0x01;		/* default not so very verbose				*/
int snull_eth;


/*
 * This structure is private to each device. It is used to pass
 * packets in and out, so there is place for a packet
 */

struct snull_priv {
    struct enet_statistics stats;
    int packetlen;
    int status;
    u8 *packetdata;
    IFBP ifbp;
};

extern struct device snull_devs[];

#define	DRV_VERSION_MAJOR	0
#define	DRV_VERSION_MINOR	9

char				drv_name[] = "WVLAN47";
char				drv_runtime_name[];
static char			snap_header[] = { 0, 0, 0xAA, 0xAA, 0x03, 0x00, 0x00, 0xF8 };

IFB_STRCT 			ifb;
CFG_ID_STRCT		ltv;			//scratch LTV-structure with value space of 17
//CFG_MAC_ADDR_STRCT	mac_addr_ltv;	//MAC Address retrieval
//CFG_..._...._STRCT	port_type_ltv;	//port type
char ssid_string[] = "Lucent Technologies";

//int		int_num = 0x0A;		//Interrupt number default
//int		io_base = 0x1200;	//I/O base default
unsigned int	major = 0;		//major number of device

#if 0 //too much work late at night
void my_printk( int level, ... );

void my_printk( int level, char *str ) {

    if ( level <= dev ) printk( str );
} /* my_print */
#endif //0        
/*
 * The "change_mtu" method is usually not needed.
 * If you need it, it must be like this.
 */
int wvlan_change_mtu(struct device *dev, int new_mtu) {

    printk( "<1>wvlan_change_mtu\n" );
    /* check ranges */
    if ((new_mtu < 68) || (new_mtu > 1500))
        return -EINVAL;
    /*
     * Do anything you need, and the accept the value
     */
    dev->mtu = new_mtu;
    return 0; /* success */
} /* wvlan_change_mtu */

/*
 * Configuration changes (passed on by ifconfig)
 */
int wvlan_config(struct device *dev, struct ifmap *map) {

    printk( "<1>wvlan_config\n" );

    if (dev->flags & IFF_UP) /* can't act on a running interface */
        return -EBUSY;

    /* Don't allow changing the I/O address */
    if (map->base_addr != dev->base_addr) {
        printk(KERN_WARNING "snull: Can't change I/O address\n");
        return -EOPNOTSUPP;
    }

    /* Allow changing the IRQ */
    if (map->irq != dev->irq) {
        dev->irq = map->irq;
        /* request_irq() is delayed to open-time */
    }

    /* ignore other fields */
    return 0;
} /* wvlan_config */

/*
 * Transmit a packet (low level (hw dependent) interface)
 */
void wvlan_hw_tx(char *buf, int len, struct device *dev) {


    /* I am paranoid. Ain't I? */
    if (len < sizeof(struct ethhdr) + sizeof(struct iphdr)) {
        printk( "snull: Hmm... packet too short (%i octets)\n", len);
        return;
    }

    if ( 0 ) { /* enable this conditional to skip displaying all the data */
//	printk( "<1>wvlan_hw_tx\n" );
    } else {
        int i;
        PDEBUG("snull: len is %i, RscInd = %i \n" KERN_DEBUG "data:", len, ifb.IFB_RscInd );
        printk( "<1>wvlan_hw_tx   len:%i   RscInd:%i", len, ifb.IFB_PIFRscInd );
        for (i=0/*14*/ ; i<len; i++) printk( "%02x ",buf[i]&0xFF);
        printk( "\n");
    }
	if ( buf[13] + (buf[12] << 8)  > 1500 ) {
		int i;
    	hcf_put_data( &ifb, (wci_bufp)buf, 12, HCF_PORT_0 );
		len += sizeof(snap_header);
		snap_header[1] = len & 0xFF; snap_header[0] = (char)(len>>8);
        printk( "<1>snapheader " );
        for (i=0 ; i<sizeof(snap_header); i++) printk( "%02x ",snap_header[i]&0xFF);
        printk( "\n");
		hcf_put_data( &ifb, (wci_bufp)&snap_header, sizeof(snap_header), HCF_PORT_0 ); 
    	hcf_put_data( &ifb, (wci_bufp)(buf+12), len - 12 - sizeof(snap_header), HCF_PORT_0 );
    	printk( "<1>E-II\n" );
	} else hcf_put_data( &ifb, (wci_bufp)buf, len, HCF_PORT_0 );
    printk( "<1>RscInd:%i\n", ifb.IFB_PIFRscInd );
    hcf_send( &ifb, HCF_PORT_0 );
} /* wvlan_hw_tx */

/*
 * The init function (sometimes called probe).
 * It is invoked by register_netdev()
 */
int wvlan_init(struct device *dev) {

int i, rc;
hcf_16 *p;
hcf_8  *cp;

    printk( "<1>wvlan_init\n" );

    /*
     * Make the usual checks: check_region(), probe irq, ...  -ENODEV
     * should be returned if no device found.  No resource should be
     * grabbed: this is done on open(). 
     */
    pcic_setup();	//;? the very first time fails,
                        //;? besides there should be some REAL error check here
    for ( i = 1000; i--; )			
        hcf_connect( &ifb, port ); //waste some time and hope for the better
    rc = check_region( ifb.IFB_IOBase, ifb.IFB_IORange );
    if ( rc < 0 ) {
        printk( "<1>port at %04x NOT reachable, continue anyway\n", port );
        rc = 0;	//;? to force execution to continue for the time being
    }
    if ( rc >= 0 ) {
        request_region( ifb.IFB_IOBase, ifb.IFB_IORange, drv_name ); //;?which name
        hcf_action( &ifb, HCF_ACT_CARD_IN );
    }
    if ( rc >= 0 ) {	//rc is here still "Linux" style
//        hcf_action( &ifb, HCF_ACT_INT_ON );
	ltv.len = 2; ltv.typ = CFG_CNF_PORT_TYPE; ltv.id[0] = port_type;
	rc = hcf_put_info( &ifb, (LTVP)&ltv );
	printk( "<1>put_info port_type = %04x   rc = %04x\n", port_type, rc );
	
	for ( cp = &ltv.id[1], i = 0; i < sizeof(ssid_string); i++ )
	    *cp++ = ssid_string[i];
	ltv.id[0] = --i;	    
	ltv.len = 18;
	ltv.typ = CFG_CNF_DESIRED_SSID;
        rc = hcf_put_info( &ifb, (LTVP)&ltv );
	printk( "<1>put_info SSID: rc = %04x\n", rc );
	
        rc = hcf_enable( &ifb, HCF_PORT_0 );
        printk( "<1>enable: rc = %04x\n", rc );
        p = (hcf_16*)&ifb;
	if ( 0 ) { //enable to see the IFB
            do {
	        printk(  "<1>" );
    	        for ( i = 0; i < 0x0010; i++ ) 
		    printk( "%04x ", *p++ );
		    printk( "\n" );
	    } while ( p < (unsigned short*)&ifb.IFB_TimStat );
	}
    }
    
    
    
    
    if ( rc == HCF_SUCCESS ) {	//rc is now "WCI" style
	hcf_action( &ifb, HCF_ACT_TICK_ON );
#if 10
	ltv.len = 4;
	ltv.typ = CFG_CNF_OWN_MAC_ADDR;
	rc = hcf_get_info( &ifb, (LTVP)&ltv );
	printk( "<1>get_info OWN_MAC_ADDR: rc = %04x\t", rc );
    	p = (hcf_16*)&ltv;
    	for ( i = 0; i < 5; i++ ) printk( "%04x ", *p++ );
	printk( "\n" );
#endif //0
    }
//    if ( rc ) cleanup_module();
//    return rc;
//} // init_module

    /* 
     * Then, assign other fields in dev, using ether_setup() and some
     * hand assignments
     */
    ether_setup(dev); /* assign some of the fields */

    dev->open            = wvlan_open;
    dev->stop            = wvlan_release;
    dev->set_config      = wvlan_config;
    dev->hard_start_xmit = wvlan_tx;
    dev->do_ioctl        = wvlan_ioctl;
    dev->get_stats       = wvlan_stats;
    dev->change_mtu      = wvlan_change_mtu;  
//    dev->rebuild_header = wvlan_rebuild_header;
    /* keep the default flags, just add NOARP */
    //dev->flags           |= IFF_NOARP;

    /*
     * Then, allocate the priv field. This encloses the statistics
     * and a few private fields.
     */
    dev->priv = kmalloc(sizeof(struct snull_priv), GFP_KERNEL);
    if (dev->priv == NULL)
        return -ENOMEM;
    memset(dev->priv, 0, sizeof(struct snull_priv));
        ((struct snull_priv*)dev->priv)->ifbp = &ifb;
    return 0;
} /* wvlan_init */

/*
 * Ioctl commands 
 */
int wvlan_ioctl(struct device *dev, struct ifreq *rq, int cmd) {

    printk( "<1>wvlan_ioctl\n" );
 
    PDEBUG("ioctl\n");
    return 0;
} /* wvlan_ioctl */

/*
 * The typical interrupt entry point (1.3.70 and later)
 */
void wvlan_isr(int irq, void *dev_id, struct pt_regs *regs) {

static int cnt = 4; //just to prove to the operator that interrupts seem to function
int len, rc;

    /*
     * As usual, check the "device" pointer for shared handlers.
     * Then assign "struct device *dev"
     */
#if 0
    /* This is the way to do things for non-shared handlers */
    struct device *dev = (struct device *)(irq2dev_map[irq]);
#else
    /* Otherwise use this SA_SHIRQ-safe approach */
    struct device *dev = (struct device *)dev_id;
    /* ... and check with hw if it's really ours */
#endif
    printk( "<1>wvlan_isr\n" );
    hcf_action( &ifb, HCF_ACT_INT_OFF );
    rc = hcf_service_nic( &ifb );
    printk( "<1>ISR status: %04X\n", rc );
//    if ( cnt && rc & HREG_EV_TICK && !--cnt ) hcf_action( &ifb, HCF_ACT_TICK_OFF );
    
//;?    if (!dev /*paranoid*/ ) return;
    dev->interrupt = 1; /* lock */

    /* retrieve statusword: real netdevices use inb() or inw() */
    len = ifb.IFB_RxLen;
    if ( len ) {	//    if (statusword & SNULL_RX_INTR)
//        printk( "<1>ISR: Rx len = %04x\n", len );
        /* send it to wvlan_rx for handling */
        wvlan_rx(dev, len );
    }
//  if (statusword & SNULL_TX_INTR) {
    if ( ifb.IFB_PIFRscInd && dev->tbusy == 1 ) {
//        /* a transmission is over: tell we are no more busy */
//        privptr->stats.tx_packets++;
        dev->tbusy = 0;	//;?feels like race condition problem, see pg 189
//	printk( "<1>wvlan_isr, tx, mark_bh(NET_BH)\n" );
        mark_bh(NET_BH);
    }
    hcf_action( &ifb, HCF_ACT_INT_ON );
    printk( "<1>IntOffCnt: %04X    IntEnMask: %04X\n",
	   ifb.IFB_IntOffCnt, ifb.IFB_IntEnMask );
    dev->interrupt = 0; /* release lock */


    return;
} /* wvlan_isr */

/*
 * Open and close
 */

int wvlan_open(struct device *dev) {

int	i, rc;
hcf_16	*p;
char	*cp;

    printk( "<1>wvlan_open\n" );

    /* request_region(), request_irq(), ....  (like fops->open) */

#if 0
    /*
     * We have no irq line, otherwise this assignment can be used to
     * grab a non-shared interrupt. To share interrupt lines use
     * the dev_id argument of request_irq. Seel wvlan_isr below.
     */
    irq2dev_map[dev->irq] = dev;
#endif
		
    rc = request_irq( irq, wvlan_isr, SA_INTERRUPT, drv_name, dev );
    printk(  "<1>request irq, rc = %04x\n", rc );
    rc = hcf_service_nic( &ifb );
    printk( "<1>NIC status: %04X\n", rc );
    rc = hcf_action( &ifb, HCF_ACT_INT_ON );
    printk( "<1>ACT_INT_ON status: %04X\n", rc );

    /* 
     * Assign the hardware address of the board: use "\0SNULx", where
     * x is 0 or 1. The first byte is '\0': a safe choice with regard
     * to multicast
     */
//    for (i=0; i < ETH_ALEN; i++)
//        dev->dev_addr[i] = "\0SNUL0"[i];
//    dev->dev_addr[ETH_ALEN-1] += (dev - snull_devs); /* the number */

    ltv.len = 4;
    ltv.typ = CFG_CNF_OWN_MAC_ADDR;
    rc = hcf_get_info( &ifb, (LTVP)&ltv );
    printk(  "<1>get_info OWN_MAC_ADDR: rc = %04x\t", rc );
    p = (hcf_16*)&ltv;
    for ( i = 0; i < 5; i++ ) printk( "%04x ", *p++ );
    printk( "\n" );
    cp = (char*)ltv.id;
    for (i=0; i < ETH_ALEN; i++)
        dev->dev_addr[i] = cp[i];



    
    ltv.len = 18;
    ltv.typ = 0xFD41; //CFG_CURRENT_SSID
    rc = hcf_get_info( &ifb, (LTVP)&ltv );
    printk( "<1>get_info SSID: rc = %04x\t", rc );
    p = (hcf_16*)&ltv;
    for ( i = 0; i < 8; i++ ) printk( "%04x ", *p++ );
    cp = (char*)&ltv.id[0];
    for ( i = 0; i < 16; i++ ) printk( "%c", (*cp++)&0xFF );
    printk( "\n" );
    


    
    ltv.len = 18;
    ltv.typ = 0xFD42; //CFG_CURRENT_BSSID
    rc = hcf_get_info( &ifb, (LTVP)&ltv );
    printk( "<1>get_info BSSID: rc = %04x\t", rc );
    p = (hcf_16*)&ltv;
    for ( i = 0; i < 8; i++ ) printk( "%04x ", *p++ );
    cp = (char*)&ltv.id[0];
    for ( i = 0; i < 16; i++ ) printk( "%c", (*cp++)&0xFF );
    printk( "\n" );
    


    dev->start = 1;
    dev->tbusy = 0;
    MOD_INC_USE_COUNT;
    return 0;
} /* wvlan_open */

/*
 * This function is called to fill up an eth header, since arp is not
 * available on the interface
 */
#if 0 //;?apparently not needed if we use ARP
int wvlan_rebuild_header(void *buff, struct device *dev, unsigned long dst,
		    struct sk_buff *skb) {
		    
struct ethhdr *eth = (struct ethhdr *)buff;

    printk(  "<1>wvlan_rebuild_header\n" );

//    memcpy(eth->h_source, dev->dev_addr, dev->addr_len);
//    memcpy(eth->h_dest, dev->dev_addr, dev->addr_len);
//    eth->h_dest[ETH_ALEN-1]   ^= 0x01;   /* dest is us xor 1 */
    return 0;
} /* wvlan_rebuild_header */
#endif //;?apparently not needed if we use ARP


int wvlan_release(struct device *dev) {
    printk( "<1>wvlan_release\n" );

    /* release ports, irq and such -- like fops->close */

    release_region( port, ifb.IFB_IORange );
    printk( "<1>release_region %04X-%04X", port, ifb.IFB_IORange );
    free_irq( irq, dev );
       
	
    dev->start = 0;
    dev->tbusy = 1; /* can't transmit any more */
    MOD_DEC_USE_COUNT;
    /* if irq2dev_map was used, zero the entry here */
    return 0;
} /* wvlan_release */

/*
 * Receive a packet: retrieve, encapsulate and pass over to upper levels
 */
void wvlan_rx(struct device *dev, int len ) {

struct	sk_buff *skb;
struct	snull_priv *privp = (struct snull_priv *)dev->priv;
char	*p;
int	i;
    /*
     * The packet has been retrieved from the transmission
     * medium. Build an skb around it, so upper layers can handle it
     */

//    printk( "<1>wvlan_rx\n" );
    skb = dev_alloc_skb(len+2);
    if (!skb) {
        printk("snull rx: low on mem\n");
        return;
    }
    skb_reserve(skb, 2); /* align IP on 16B boundary */  
    p = skb_put(skb, len);
    printk( "<1>wvlan_rx   len:%i   stat:%04x", ifb.IFB_RxLen, ifb.IFB_RxStat );
	if (  ifb.IFB_RxStat == 0x2000 || ifb.IFB_RxStat == 0x4000 ) {
    	hcf_get_data( &ifb, 0, p /*skb_put( skb, len )*/, 12 );
    	hcf_get_data( &ifb, 20, p+12 /*skb_put( skb, len )*/, len-12 );
	} else hcf_get_data( &ifb, 0, p /*skb_put( skb, len )*/, len );
    if ( 1 ) { /* enable this conditional to skip displaying all the data */
//        printk( "<1>wvlan_rx\n" );
    } else {
        printk( "<1>rcvd data %d :", len );
	for ( i = len; i; i-- ) printk( "%02X ", (*p++)&0xFF );
        printk( "\n" );
    }
    /* Write metadata, and then pass to the receive level */
    skb->dev = dev;
    skb->protocol = eth_type_trans(skb, dev);
    skb->ip_summed = CHECKSUM_UNNECESSARY; /* don't check it */
    privp->stats.rx_packets++;
    netif_rx(skb);
    return;
} /* wvlan_rx */
    
/*
 * Return statistics to the caller
 */
struct enet_statistics *wvlan_stats(struct device *dev) {

    struct snull_priv *priv = (struct snull_priv *)dev->priv;
    printk( "<1>enet_statistics\n" );

    return &priv->stats;
} /* wvlan_stats */

/*
 * Transmit a packet (called by the kernel)
 */
int wvlan_tx(struct sk_buff *skb, struct device *dev) {

int len, retval=0;
char *cp, *data;
hcf_16 *p;
int i, rc;

//    printk( "<1>wvlan_tx\n" );

    
    ltv.len = 18;
    ltv.typ = 0xFD42; //CFG_CURRENT_BSSID
    rc = hcf_get_info( &ifb, (LTVP)&ltv );
//    printk( "<1>get_info SSID: rc = %04x\t", rc );
    p = (hcf_16*)&ltv;
//    for ( i = 0; i < 8; i++ ) printk( "%04x ", *p++ );
    cp = (char*)&ltv.id[0];
//    for ( i = 0; i < 16; i++ ) printk(  "%c", (*cp++)&0xFF );
//    printk( "\n" );

    
    ltv.len = 18;
    ltv.typ = 0xFD41; //CFG_CURRENT_SSID
    rc = hcf_get_info( &ifb, (LTVP)&ltv );
//    printk( "<1>get_info SSID: rc = %04x\t", rc );
    p = (hcf_16*)&ltv;
//    for ( i = 0; i < 8; i++ ) printk(  "%04x ", *p++ );
    cp = (char*)&ltv.id[0];
//    for ( i = 0; i < 16; i++ ) printk( "%c", (*cp++)&0xFF );
//    printk(  "\n" );
    

    

    if (dev->tbusy) {/* shouldn't happen */
        printk( "<1>wvlan_tx called while tbusy != 0\n" );
        retval = -EBUSY;
//        goto tx_done;
//;?	changed based on recomandation of Andreas. once it works I should figure this out
//	+1 is unexpected, I would have guessed something like EPERM, but than all errors seems
//	to be negated
//        retval = 1; //;?EPERM
        goto tx_L99;
    }
    if (skb == NULL) {
        PDEBUG("tint for %p\n",dev);
        dev_tint(dev); /* we are ready to transmit */
        return 0;
    }
    dev->tbusy = 1; /* transmission is busy */

    len = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN; /* minimum len */
    data = skb->data;
    dev->trans_start = jiffies; /* save the timestamp */

    /* actual deliver of data is device-specific, and not shown here */
    wvlan_hw_tx(data, len, dev);

  tx_done:
    dev_kfree_skb(skb, FREE_WRITE); /* release it */
  tx_L99:
    return retval; /* zero == done; nonzero == fail */
} /* wvlan_tx */

void pcic_pair_out( hcf_16 reg, hcf_8 val ) {

    outb( reg, PCIC_PORT );
  printk(  "<1>%04x: %02x  ", PCIC_PORT, inb_p( PCIC_PORT ) );
    outb( val, PCIC_PORT+1 );
  printk(  "   %04x: %02x\n", PCIC_PORT+1, inb_p( PCIC_PORT+1 ) );
} // pcic_pair_out

void pcic_setup( void ) {

int i, rc;
unsigned long addr;

    rc = check_region( PCIC_PORT, PCIC_RANGE );
    if ( rc < 0 ) {
	printk(  "<1>PCIC at %04x NOT reachable, continue anyway\n", PCIC_PORT );
//	rc = 0;	//;? to force execution to continue for the time being
    }
    request_region( PCIC_PORT, PCIC_RANGE, drv_name );
    pcic_pair_out( 0x0002, 0x0000 );
    pcic_pair_out( 0x0002, 0x0010 );
//    pcic_pair_out( 0x0003, 0x0060 );
    pcic_pair_out( 0x0002, 0x0090 );
    pcic_pair_out( 0x0010, 0x00D0 );
    pcic_pair_out( 0x0011, 0x0000 );
    pcic_pair_out( 0x0012, 0x00D1 );
    pcic_pair_out( 0x0013, 0x0040 );
    pcic_pair_out( 0x0014, 0x0030 );
    pcic_pair_out( 0x0015, 0x007F );
    pcic_pair_out( 0x0006, 0x0001 );
    /*;? next line is apparently not sufficient to get anything but IRQ 0x10 */
    pcic_pair_out( 0x0003, 0x0040 );		//drop reset bit
    pcic_pair_out( 0x0003, 0x0060 );		//add I/O mode bit
    pcic_pair_out( 0x0003, 0x0060 + irq );		//0x60 + Int.vector
    pcic_pair_out( 0x0008, (hcf_8)port );		//IO_base low
    pcic_pair_out( 0x0009, (hcf_8)(port>>8) );		//IO_base high
    pcic_pair_out( 0x000A, (hcf_8)(port + 0x0040) );	//IO_end  low
    pcic_pair_out( 0x000B, (hcf_8)(port>>8) );		//IO_end  high
    pcic_pair_out( 0x0007, 0x0009 );			//IO window 0, 16 bit, DeusExMachina bit
    pcic_pair_out( 0x0006, 0x0041 );			//enable memory window
    //COR Access
    addr = 0xD0054;
    
    do {
	printk(   "<1>" );
	for ( i = 0; i < 0x0010; i++, addr += 2 ) 
	    printk(  "%02x ", *(unsigned char*)addr );
	printk( "     " );
	addr -= 0x0020;
	for ( i = 0; i < 0x0010; i++, addr += 2 ) 
	    printk(  "%c", *(unsigned char*)addr );
        printk(  "\n" );
    } while ( addr < 0xD0094 );
    addr = 0xD03E0;			//set COR
    *(unsigned char*)addr = 0x80;
    for ( i = 0; i < 65000; i++ ) /*NOP*/;//    delay
    *(unsigned char*)addr = 0x00;
    for ( i = 0; i < 65000; i++ ) /*NOP*/;//    delay
    *(unsigned char*)addr = 0x45;
    pcic_pair_out( 0x0006, 0x0040 );	//disable memory window
    release_region( PCIC_PORT, PCIC_RANGE );
	
} /* pcic_setup */


/*
 * The devices
 */

char wvlan_names[20]; /* two 10-byte buffers */
struct device snull_devs[2] = {
    {
        wvlan_names, /* name -- set at load time */
        0, 0, 0, 0,  /* shmem addresses */
        0x000,       /* ioport */
        0,           /* irq line */
        0, 0, 0,     /* various flags: init to 0 */
        NULL,        /* next ptr */
        wvlan_init,  /* init function, fill other fields with NULL's */
    },
    {
        wvlan_names+10,/* name -- set at load time */
        0, 0, 0, 0,  /* shmem addresses */
        0x000,       /* ioport */
        0,           /* irq line */
        0, 0, 0,     /* various flags: init to 0 */
        NULL,        /* next ptr */
        wvlan_init,  /* init function, fill other fields with NULL's */
    }
};

/*
 * Finally, the module stuff
 */

int init_module(void)
{

    int result, i, device_present = 0;

    printk( "<1>init_module\n" );

    snull_eth = eth; /* copy the cfg datum in the non-static place */

    if (!snull_eth) { /* call them "wvlan47#" */
        memcpy(snull_devs[0].name, "wvlan470", 8);
        memcpy(snull_devs[1].name, "wvlan471", 8);
    } else { /* use automatic assignment (eth#) */
        snull_devs[0].name[0] = snull_devs[1].name[0] = ' ';
    }

    for (i=0; i<1 /*2*/;  i++)	//;?think this over later when you are ready for multiple card support
        if ( (result = register_netdev(snull_devs + i)) )
            printk( "wvlan47: error %i registering device \"%s\"\n",
                   result, snull_devs[i].name);
        else device_present++;
#ifndef SNULL_DEBUG
    register_symtab(NULL); /* hide symbols */
#endif

    return device_present ? 0 : -ENODEV;
}

void cleanup_module(void)
{
    int i;
    printk( "<1>cleanup_module\n" );

    for (i=0; i<1 /*;?2*/;  i++) {
        kfree(snull_devs[i].priv);
        unregister_netdev(snull_devs + i);
    }
    return;
}






